'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

# Test suites:
from webuntis.tests.offline import *
#from webuntis.tests.online import *
